/*  Copyright (c) MediaArea.net SARL & AV Preservation by reto.ch.
*
*  Use of this source code is governed by a BSD-style license that can
*  be found in the License.html file in the root of the source tree.
*/

//---------------------------------------------------------------------------
#ifndef License_ListH
#define License_ListH
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#include "Lib/DPX/DPX.h"
#include "Lib/TIFF/TIFF.h"
#include "Lib/WAV/WAV.h"
#include "Lib/AIFF/AIFF.h"
#include "Lib/License.h"
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// Flags map
const uint8_t FlagsMap[16] =
{
     0,
     3 + Parser_DPX,
     3 + Parser_WAV,
     3 + Parser_TIFF,
     3 + Parser_AIFF,
    (uint8_t)-1,
     1,
     2,
    (uint8_t)-1,
    (uint8_t)-1,
    (uint8_t)-1,
    (uint8_t)-1,
    (uint8_t)-1,
    (uint8_t)-1,
    (uint8_t)-1,
    (uint8_t)-1,
};

//---------------------------------------------------------------------------
// Flags - Features
static const uint8_t Features_Flags_List[Feature_Max + 1] =
{
    Feature_GeneralOptions,
    Feature_InputOptions,
    Feature_EncodingOptions,
    Feature_MultipleTracks,
    (uint8_t)-1,
};

//---------------------------------------------------------------------------
// Flags - Muxers
static const uint8_t Muxers_Flags_List[Muxer_Max + 1] =
{
    Muxer_Matroska,
    (uint8_t)-1,
};

//---------------------------------------------------------------------------
// Flags - Encoders
static const uint8_t Encoders_Flags_List[Encoder_Max + 1] =
{
    Encoder_PCM,
    Encoder_FFV1,
    Encoder_FLAC,
    (uint8_t)-1,
};

//---------------------------------------------------------------------------
// Flags - Parsers - DPX
static const uint8_t DPX_Flags_List[dpx::Style_Max + 1] =
{
    dpx::Raw_RGBA_10_FilledA_LE,
    dpx::Raw_RGBA_10_FilledA_BE,
    dpx::Raw_RGB_16_BE,
    dpx::Raw_RGB_16_LE,
    dpx::Raw_RGBA_16_BE,
    dpx::Raw_RGBA_16_LE,
    dpx::Raw_RGB_8,
    dpx::Raw_RGBA_8,
    dpx::Raw_RGB_10_FilledA_LE,
    dpx::Raw_RGB_10_FilledA_BE,
    dpx::Raw_RGB_12_Packed_BE,
    dpx::Raw_RGB_12_FilledA_BE,
    dpx::Raw_RGB_12_FilledA_LE,
    dpx::Raw_RGBA_12_Packed_BE,
    dpx::Raw_RGBA_12_FilledA_BE,
    dpx::Raw_RGBA_12_FilledA_LE,
    (uint8_t)-1,
};

//---------------------------------------------------------------------------
// Flags - Parsers - WAV
static const uint8_t WAV_Flags_List[wav::Style_Max + 1] =
{
    wav::PCM_48000_24_1_LE,
    wav::PCM_48000_24_2_LE,
    wav::PCM_48000_24_6_LE,
    wav::PCM_96000_24_1_LE,
    wav::PCM_96000_24_2_LE,
    wav::PCM_96000_24_6_LE,
    wav::PCM_48000_16_1_LE,
    wav::PCM_48000_16_2_LE,
    wav::PCM_48000_16_6_LE,
    wav::PCM_96000_16_1_LE,
    wav::PCM_96000_16_2_LE,
    wav::PCM_96000_16_6_LE,
    wav::PCM_48000_8_1_U,
    wav::PCM_48000_8_2_U,
    wav::PCM_48000_8_6_U,
    wav::PCM_96000_8_1_U,
    wav::PCM_96000_8_2_U,
    wav::PCM_96000_8_6_U,
    wav::PCM_44100_24_1_LE,
    wav::PCM_44100_24_2_LE,
    wav::PCM_44100_24_6_LE,
    wav::PCM_44100_16_1_LE,
    wav::PCM_44100_16_2_LE,
    wav::PCM_44100_16_6_LE,
    wav::PCM_44100_8_1_U,
    wav::PCM_44100_8_2_U,
    wav::PCM_44100_8_6_U,
    (uint8_t)-1,
};

//---------------------------------------------------------------------------
// Flags - Parsers - TIFF
static const uint8_t TIFF_Flags_List[tiff::Style_Max + 1] =
{
    tiff::Raw_RGB_16_U_BE,
    tiff::Raw_RGB_16_U_LE,
    tiff::Raw_RGBA_16_U_LE,
    tiff::Raw_RGB_8_U,
    tiff::Raw_RGBA_8_U,
    (uint8_t)-1,
};

//---------------------------------------------------------------------------
// Flags - Parsers - AIFF
static const uint8_t AIFF_Flags_List[aiff::Style_Max + 1] =
{
    aiff::PCM_48000_24_1_BE,
    aiff::PCM_48000_24_2_BE,
    aiff::PCM_48000_24_6_BE,
    aiff::PCM_96000_24_1_BE,
    aiff::PCM_96000_24_2_BE,
    aiff::PCM_96000_24_6_BE,
    aiff::PCM_48000_16_1_BE,
    aiff::PCM_48000_16_1_LE,
    aiff::PCM_48000_16_2_BE,
    aiff::PCM_48000_16_2_LE,
    aiff::PCM_48000_16_6_BE,
    aiff::PCM_48000_16_6_LE,
    aiff::PCM_96000_16_1_BE,
    aiff::PCM_96000_16_1_LE,
    aiff::PCM_96000_16_2_BE,
    aiff::PCM_96000_16_2_LE,
    aiff::PCM_96000_16_6_BE,
    aiff::PCM_96000_16_6_LE,
    aiff::PCM_44100_8_1_S,
    aiff::PCM_44100_8_1_U,
    aiff::PCM_44100_8_2_S,
    aiff::PCM_44100_8_2_U,
    aiff::PCM_44100_8_6_S,
    aiff::PCM_44100_8_6_U,
    aiff::PCM_44100_16_1_BE,
    aiff::PCM_44100_16_1_LE,
    aiff::PCM_44100_16_2_BE,
    aiff::PCM_44100_16_2_LE,
    aiff::PCM_44100_16_6_BE,
    aiff::PCM_44100_16_6_LE,
    aiff::PCM_44100_24_1_BE,
    aiff::PCM_44100_24_2_BE,
    aiff::PCM_44100_24_6_BE,
    aiff::PCM_48000_8_1_S,
    aiff::PCM_48000_8_1_U,
    aiff::PCM_48000_8_2_S,
    aiff::PCM_48000_8_2_U,
    aiff::PCM_48000_8_6_S,
    aiff::PCM_48000_8_6_U,
    aiff::PCM_96000_8_1_S,
    aiff::PCM_96000_8_1_U,
    aiff::PCM_96000_8_2_S,
    aiff::PCM_96000_8_2_U,
    aiff::PCM_96000_8_6_S,
    aiff::PCM_96000_8_6_U,
    (uint8_t)-1,
};

static const uint8_t* Flags_Lists[3 + Parser_Max + 1] =
{
    Features_Flags_List,
    Muxers_Flags_List,
    Encoders_Flags_List,
    DPX_Flags_List,
    TIFF_Flags_List,
    WAV_Flags_List,
    AIFF_Flags_List,
    NULL,
};

//---------------------------------------------------------------------------
// Default license content
struct default_license_values
{
    uint8_t         Value;
    uint8_t         Flavor;
};
static default_license_values DefaultLicense_Parsers[] =
{
    { 1                 , Muxer_Matroska                },
    { 2                 , Encoder_FFV1                  },
    { 2                 , Encoder_FLAC                  },
    { 3 + Parser_DPX    , dpx::Raw_RGB_8                },
    { 3 + Parser_DPX    , dpx::Raw_RGB_10_FilledA_LE    },
    { 3 + Parser_DPX    , dpx::Raw_RGB_10_FilledA_BE    },
    { 3 + Parser_WAV    , wav::PCM_48000_16_2_LE        },
    { (uint8_t)-1       , (uint8_t)-1                   },
};

//---------------------------------------------------------------------------
struct feature_info
{
    const char*     Name;
};
const feature_info FeatureInfos[Feature_Max + 1] =
{
    { "General options"         },
    { "Input options"           },
    { "Encoding options"        },
    { "More than 2 tracks"      },
    { NULL                      },
};

//---------------------------------------------------------------------------
struct muxer_info
{
    const char*     Name;
};
const muxer_info MuxerInfos[Muxer_Max] =
{
    { "Matroska"    },
};

//---------------------------------------------------------------------------
struct encoder_info
{
    const char*     Name;
};
const encoder_info EncoderInfos[Encoder_Max] =
{
    { "FFV1"        },
    { "PCM"         },
    { "FLAC"        },
};

//---------------------------------------------------------------------------
static string Features_String(uint8_t Value) { return FeatureInfos[Value].Name; }
static string Muxers_String(uint8_t Value) { return MuxerInfos[Value].Name; }
static string Encoders_String(uint8_t Value) { return EncoderInfos[Value].Name; }
typedef string (*flavor_string) (uint8_t Flavor);
struct parser_info
{
    const char*     Name;
    uint8_t         Size;
    flavor_string   Flavor_String;
};
static const parser_info Infos[3 + Parser_Max + 1] =
{
    { "features"            , Feature_Max           , Features_String       },
    { "muxers/demuxers"     , Muxer_Max             , Muxers_String         },
    { "encoders/decoders"   , Encoder_Max           , Encoders_String       },
    { "DPX flavors"         , dpx::Style_Max        , DPX_Flavor_String     },
    { "TIFF flavors"        , tiff::Style_Max       , TIFF_Flavor_String    },
    { "WAV flavors"         , wav::Style_Max        , WAV_Flavor_String     },
    { "AIFF flavors"        , aiff::Style_Max       , AIFF_Flavor_String    },
    { NULL                  , 0                     , NULL                  },
};

//---------------------------------------------------------------------------
struct license_internal
{
    time_t              Date;
    uint64_t            UserID;
    uint64_t            License_Flags[3 + Parser_Max];

    license_internal();

    static void         AddFlavor(uint64_t &Value, uint8_t Flavor);
    static const char*  SupportedFlavor(uint64_t &Value, uint8_t Flavor);

    // License info
    uint8_t             Input_Flags[3];
};

#endif
